// Copyright 2000-2019 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.diagnostic;

import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.util.Comparing;
import gnu.trove.THashSet;
import gnu.trove.TObjectHashingStrategy;
import org.jetbrains.annotations.ApiStatus;
import org.jetbrains.annotations.NotNull;

import java.util.Arrays;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;
import java.util.stream.Collectors;

@ApiStatus.Internal
public enum LoadingPhase {
  BOOTSTRAP,
  SPLASH,
  CONFIGURATION_STORE_INITIALIZED,
  FRAME_SHOWN,
  PROJECT_OPENED,
  INDEXING_FINISHED;

  @NotNull
  private static Logger getLogger() {
    return Logger.getInstance(LoadingPhase.class);
  }

  private final static boolean KEEP_IN_MIND_LOADING_PHASE = Boolean.parseBoolean("idea.keep.in.mind.loading.phase");

  public static void setCurrentPhase(@NotNull LoadingPhase phase) {
    currentPhase.set(phase);
    logPhaseSet(phase);
  }

  private final static Set<Throwable> stackTraces = new THashSet<>(new TObjectHashingStrategy<Throwable>() {
    @Override
    public int computeHashCode(Throwable throwable) {
      return getCollect(throwable).hashCode();
    }

    private String getCollect(Throwable throwable) {
      return Arrays
        .stream(throwable.getStackTrace())
        .map(element -> element.getClassName() + element.getMethodName())
        .collect(Collectors.joining());
    }

    @Override
    public boolean equals(Throwable o1, Throwable o2) {
      if (o1 == o2) return true;
      if (o1 == null || o2 == null) return false;
      return Comparing.equal(getCollect(o1), getCollect(o2));
    }
  });

  private final static AtomicReference<LoadingPhase> currentPhase = new AtomicReference<>(BOOTSTRAP);

  public static void compareAndSet(@NotNull LoadingPhase expect, @NotNull LoadingPhase phase) {
    if (currentPhase.compareAndSet(expect, phase)) {
      logPhaseSet(phase);
    }
  }

  private static void logPhaseSet(@NotNull LoadingPhase phase) {
    getLogger().info("Reached " + phase + " loading phase");
  }

  public static void assertAtLeast(@NotNull LoadingPhase phase) {
    if (!KEEP_IN_MIND_LOADING_PHASE) {
      return;
    }

    LoadingPhase currentPhase = LoadingPhase.currentPhase.get();
    if (currentPhase.ordinal() >= phase.ordinal() || isKnownViolator()) {
      return;
    }

    Throwable t = new Throwable();
    synchronized (stackTraces) {
      if (!stackTraces.add(t)) {
        return;
      }

      getLogger().warn("Should be called at least at phase " + phase + ", the current phase is: " + currentPhase + "\n" +
                       "Current violators count: " + stackTraces.size() + "\n\n",
                       t);
    }
  }

  private static boolean isKnownViolator() {
    return Arrays.stream(Thread.currentThread().getStackTrace())
      .anyMatch(element -> {
        String className = element.getClassName();
        return className.equals("com.intellij.openapi.application.Preloader") ||
               className.contains("com.intellij.util.indexing.IndexInfrastructure");
      });
  }

  public static boolean isStartupComplete() {
    return isComplete(INDEXING_FINISHED);
  }

  public static boolean isComplete(@NotNull LoadingPhase phase) {
    return currentPhase.get().ordinal() >= phase.ordinal();
  }
}
